import sympy as sp

from .utils import is_invertible, of_arity
from .vectorspace import VectorSpace


class LinearMapError(Exception):
    def __init__(self, msg=''):
        super().__init__(msg)


class IsomorphismError(Exception):
    def __init__(self, msg=''):
        super().__init__(msg)


class LinearMap:
    """
    pass
    """
    
    def __init__(self, domain, codomain, mapping=None, matrix=None, name=None):
        """
        pass

        Parameters
        ----------
        domain : VectorSpace
            The domain of the linear map.
        codomain : VectorSpace
            The codomain of the linear map.
        mapping : callable, optional
            A function that takes a vector in the domain and returns a 
            vector in the codomain.
        matrix : list of list or sympy.Matrix, optional
            The matrix representation of the linear map with respect to 
            the basis vectors of the domain and codomain.
        name : str, optional
            pass

        Returns
        -------
        LinearMap
            pass

        Raises
        ------
        LinearMapError
            If neither the mapping nor the matrix is provided.
        LinearMapError
            If the field of the domain and codomain are not the same.
        """
        if not isinstance(domain, VectorSpace):
            raise TypeError('The domain must be a VectorSpace.')
        if not isinstance(codomain, VectorSpace):
            raise TypeError('The codomain must be a VectorSpace.')
        if mapping is None and matrix is None:
            raise LinearMapError('Either a matrix or mapping must be provided.')
        if domain.field is not codomain.field:
            raise LinearMapError(
                'The domain and codomain must be vector spaces over the same field.'
                )
        
        if mapping is None:
            mapping = LinearMap._from_matrix(domain, codomain, matrix)
        elif not of_arity(mapping, 1):
            raise TypeError('Mapping must be a function of arity 1.')
        if matrix is None:
            matrix = LinearMap._to_matrix(domain, codomain, mapping)
        else:
            matrix = sp.Matrix(matrix)
        
        self._domain = domain
        self._codomain = codomain
        self._mapping = mapping
        self._matrix = matrix
        if name is not None:
            self.__name__ = name

    @staticmethod
    def _to_matrix(domain, codomain, mapping):
        matrix = []
        for vec in domain.basis:
            mapped_vec = mapping(vec)
            coord_vec = codomain.to_coordinate(mapped_vec)
            matrix.extend(coord_vec)
        return sp.Matrix(domain.dim, codomain.dim, matrix).T

    @staticmethod
    def _from_matrix(domain, codomain, matrix):
        matrix = sp.Matrix(matrix)
        def to_coord(vec): return sp.Matrix(domain.to_coordinate(vec))
        def from_coord(vec): return codomain.from_coordinate(vec.flat())
        return lambda vec: from_coord(matrix @ to_coord(vec))

    @property
    def field(self):
        """
        {Real, Complex}: The field of the domain and codomain.
        """
        return self.domain.field

    @property
    def domain(self):
        """
        VectorSpace: The domain of the linear map.
        """
        return self._domain
    
    @property
    def codomain(self):
        """
        VectorSpace: The codomain of the linear map.
        """
        return self._codomain
    
    @property
    def mapping(self):
        """
        callable: The function that maps vectors from the domain to the codomain.
        """
        return self._mapping
    
    @property
    def matrix(self):
        """
        sympy.Matrix: The matrix representation of the linear map.
        """
        return self._matrix
    
    @property
    def rank(self):
        """
        int: The rank of the linear map.
        """
        return self.matrix.rank()
    
    @property
    def nullity(self):
        """
        int: The nullity of the linear map.
        """
        return self.matrix.cols - self.rank
    
    def __repr__(self):
        return (
            f'LinearMap(domain={self.domain}, '
            f'codomain={self.codomain}, '
            f'mapping={self.mapping.__name__}, '
            f'matrix={self.matrix})'
            )
    
    def __str__(self):
        return self.__repr__()

    def __eq__(self, map2):
        return (
            self.domain == map2.domain 
            and self.codomain == map2.codomain 
            and self.matrix == map2.matrix
            )
    
    def __add__(self, map2):
        """
        The sum of two linear maps.

        Parameters
        ----------
        map2 : LinearMap
            The linear map being added.

        Returns
        -------
        LinearMap
            The sum of `self` and `map2`.

        Examples
        --------
        
        >>> R3 = fn(Real, 3)
        >>> def mapping1(vec): return [2*i for i in vec]
        >>> def mapping2(vec): return [3*i for i in vec]
        >>> map1 = LinearMap(R3, R3, mapping1)
        >>> map2 = LinearMap(R3, R3, mapping2)
        >>> map3 = map1 + map2
        >>> map3([1, 2, 3])
        [5, 10, 15]
        """
        # FIX: Add check to make sure the domains and codomains are equal
        def mapping(vec):
            vec1 = self.mapping(vec)
            vec2 = map2.mapping(vec)
            return self.codomain.add(vec1, vec2)
        matrix = self.matrix + map2.matrix
        return LinearMap(self.domain, self.codomain, mapping, matrix)
    
    def __mul__(self, scalar):
        """
        The product of a linear map and scalar.

        Parameters
        ----------
        scalar : Real or Complex
            The scalar to multiply with.

        Returns
        -------
        LinearMap
            The product of `self` and `scalar`.

        Examples
        --------
        
        >>> R3 = fn(Real, 3)
        >>> def mapping(vec): return [2*i for i in vec]
        >>> map1 = LinearMap(R3, R3, mapping)
        >>> map2 = 3 * map1
        >>> map2([1, 2, 3])
        [6, 12, 18]
        """
        if not isinstance(scalar, self.field):
            raise TypeError('Scalar must be an element of the vector space field.')
        def mapping(vec):
            return self.codomain.mul(scalar, self.mapping(vec))
        matrix = self.matrix * scalar
        return LinearMap(self.domain, self.codomain, mapping, matrix)
    
    def __rmul__(self, scalar):
        return self.__mul__(scalar)
    
    def __call__(self, vec):
        """
        Apply the linear map to a vector.

        Parameters
        ----------
        vec : object
            The vector to map.

        Returns
        -------
        object
            The vector that `vec` maps to.

        Examples
        --------
        
        >>> R3 = fn(Real, 3)
        >>> def mapping(vec): return [2*i for i in vec]
        >>> map1 = LinearMap(R3, R3, mapping)
        >>> map1([1, 2, 3])
        [2, 4, 6]
        """
        if vec not in self.domain:
            raise TypeError(f'{vec} is not an element of the domain.')
        return self.mapping(vec)
    
    def composition(self, map2):
        """
        The composition of two linear maps.

        Parameters
        ----------
        map2 : LinearMap
            The linear map to compose with.

        Returns
        -------
        LinearMap
            The composition of `self` and `map2`.

        Raises
        ------
        LinearMapError
            If the domain of `self` is not equal to the codomain of `map2`.

        Examples
        --------
        
        >>> R3 = fn(Real, 3)
        >>> def mapping1(vec): return [2*i for i in vec]
        >>> def mapping2(vec): return [3*i for i in vec]
        >>> map1 = LinearMap(R3, R3, mapping1)
        >>> map2 = LinearMap(R3, R3, mapping2)
        >>> map3 = map1.composition(map2)
        >>> map3([1, 2, 3])
        [6, 12, 18]
        """
        if self.domain != map2.codomain:
            raise LinearMapError('The linear maps are not compatible.')
        
        def mapping(vec):
            return self.mapping(map2.mapping(vec))
        matrix = self.matrix @ map2.matrix
        if hasattr(self, '__name__') and hasattr(map2, '__name__'):
            name = f'{self.__name__} o {map2.__name__}'
        else:
            name = None
        return LinearMap(map2.domain, self.codomain, mapping, matrix, name)
    
    def range(self):
        """
        The range, or image, of the linear map.

        Returns
        -------
        VectorSpace
            The range of `self`.

        See Also
        --------
        LinearMap.image
        """
        basis = [vec.tolist() for vec in self.matrix.columnspace()]
        basis = [self.domain.from_coordinate(vec) for vec in basis]
        return self.domain.span(*basis)

    def nullspace(self):
        """
        The null space, or kernel, of the linear map.

        Returns
        -------
        VectorSpace
            The nullspace of `self`.

        See Also
        --------
        LinearMap.kernel
        """
        basis = [vec.tolist() for vec in self.matrix.nullspace()]
        basis = [self.domain.from_coordinate(vec) for vec in basis]
        return self.domain.span(*basis)
    
    def pseudoinverse(self):
        """
        The pseudoinverse of the linear map.
        """
        raise NotImplementedError()
    
    def adjoint(self):
        """
        The adjoint of the linear map.
        """
        raise NotImplementedError()

    def is_injective(self):
        """
        Check whether the linear map is injective.

        Returns
        -------
        bool
            True if the linear map is injective, otherwise False.

        See Also
        --------
        LinearMap.is_surjective, LinearMap.is_bijective
        """
        return self.matrix.cols == self.rank

    def is_surjective(self):
        """
        Check whether the linear map is surjective.

        Returns
        -------
        bool
            True if the linear map is surjective, otherwise False.

        See Also
        --------
        LinearMap.is_injective, LinearMap.is_bijective
        """
        return self.matrix.rows == self.rank
    
    def is_bijective(self):
        """
        Check whether the linear map is bijective.

        Returns
        -------
        bool
            True if the linear map is bijective, otherwise False.

        See Also
        --------
        LinearMap.is_injective, LinearMap.is_surjective
        """
        return is_invertible(self.matrix)

    # Aliases
    image = range
    kernel = nullspace


class Isomorphism(LinearMap):
    """
    pass
    """

    def __init__(self, domain, codomain, mapping=None, matrix=None, name=None):
        super().__init__(domain, codomain, mapping, matrix, name)

        if not self.is_bijective():
            raise IsomorphismError('Linear map is not invertible.')

    def __repr__(self):
        return super().__repr__().replace('LinearMap', 'Isomorphism')
    
    def inverse(self):
        """
        The inverse of the linear map.

        Returns
        -------
        Isomorphism
            The inverse of `self`.
        """
        matrix = self.matrix.inv()
        return Isomorphism(self.codomain, self.domain, matrix=matrix)


class IdentityMap(Isomorphism):
    """
    pass
    """

    def __init__(self, vectorspace):
        """
        pass

        Parameters
        ----------
        vectorspace : VectorSpace
            The domain and codomain of the identity map.

        Returns
        -------
        IdentityMap
            pass
        """
        super().__init__(vectorspace, vectorspace, lambda vec: vec)


import sympy as sp

from .operations import InnerProduct
from .vectorspace import VectorSpace


class InnerProductSpaceError(Exception):
    def __init__(self, msg=''):
        super().__init__(msg)


class InnerProductSpace:
    """
    pass
    """

    def __init__(self, vectorspace, innerproduct=None):
        """
        pass

        Parameters
        ----------
        vectorspace : VectorSpace
            pass
        innerproduct : callable
            pass

        Returns
        -------
        InnerProductSpace
            pass
        """
        if not isinstance(vectorspace, VectorSpace):
            raise TypeError('vectorspace must be of type VectorSpace.')
        self._innerproduct = InnerProductSpace._init_innerproduct(innerproduct)

    @staticmethod
    def _init_innerproduct(ip):
        if ip is None:
            raise NotImplementedError()
        return ip
    
    def dot(self, vec1, vec2):
        """
        The dot product between two vectors.

        Parameters
        ----------
        vec1, vec2
            The vectors in the inner product space.

        Returns
        -------
        float
            The dot product between `vec1` and `vec2`.

        See Also
        --------
        VectorSpace.norm, VectorSpace.are_orthogonal

        Examples
        --------

        >>> vs = fn(Real, 3)
        >>> vs.dot([1, 2, 3], [4, 5, 6])
        32
        >>> vs.dot([1, 0, 1], [0, 1, 0])
        0
        """
        return self._innerproduct(vec1, vec2)

    def ortho_complement(self):
        """
        The orthogonal complement of an inner product space.

        Returns
        -------
        InnerProductSpace
            The orthogonal complement of `self`.

        See Also
        --------
        VectorSpace.ortho_projection, VectorSpace.dot

        Examples
        --------

        >>> vs = fn(Real, 3, constraints=['v0 == v1'])
        >>> vs.ortho_complement().basis
        [[1, -1, 0]]
        >>> vs.ortho_complement().ortho_complement() == vs
        True
        """
        raise NotImplementedError()
    
    def ortho_projection(self, vs2):
        """
        The orthogonal projection of `self` onto `vs2`.

        Parameters
        ----------
        vs2 : VectorSpace
            pass

        Returns
        -------
        InnerProductSpace
            pass

        Raises
        ------
        InnerProductSpaceError
            If `self` and `vs2` do not share the same ambient space.

        See Also
        --------
        VectorSpace.ortho_complement, VectorSpace.dot
        """
        raise NotImplementedError()


import sympy as sp


class ParsingError(Exception):
    def __init__(self, msg=''):
        super().__init__(msg)


class ConstraintError(Exception):
    def __init__(self, msg=''):
        super().__init__(msg)


def sympify(expr, allowed_vars=None):
    """
    Return the sympy representation of the given expression.

    Parameters
    ----------
    expr : str
        pass
    allowed_vars : iterable, optional
        pass

    Returns
    -------
    sympy.Basic
        pass

    Raises
    ------
    sympy.SympifyError
        pass
    ParsingError
        If `expr` contains variables not in `allowed_vars`.
    """
    # Filter unrecognized characters for safety (consider regex)
    expr = sp.sympify(expr, rational=True)  # evaluate flag
    if allowed_vars is not None:
        if not all(var in allowed_vars for var in expr.free_symbols):
            invalid_vars = expr.free_symbols - set(allowed_vars)
            raise ParsingError(f'Unrecognized variables found: {invalid_vars}')
    return expr


def split_constraint(constraint: str):
    """
    Split a constraint with multiple relational operators into separate 
    relations.

    Parameters
    ----------
    constraint : str
        pass

    Returns
    -------
    relations : set of str
        pass
    """
    exprs = constraint.split('==')
    expr_count = len(exprs)
    if expr_count == 1:
        raise ConstraintError('Constraints must include at least one "==".')
    
    eqs = set()
    for i in range(expr_count - 1):
        eq = f'{exprs[i]} - ({exprs[i + 1]})'
        eqs.add(eq)
    return eqs


try:
    import manim
except ImportError:
    raise ImportError(
        'Manim is required for the visualize module. Install it using: \n\n'
        '    pip install ablina[visualize]'
        )

from .linearmap import LinearMap
from .vectorspace import VectorSpace



import sympy as sp

from .utils import of_arity, symbols


class OperationError(Exception):
    def __init__(self, msg=''):
        super().__init__(msg)


class Operation:
    def __init__(self, func, arity):
        if not of_arity(func, arity):
            raise OperationError()
        self._func = func
        self._arity = arity

    @property
    def func(self):
        return self._func
    
    @property
    def arity(self):
        return self._arity
    
    def __call__(self, *args):
        return self.func(*args)


class VectorAdd(Operation):
    def __init__(self, field, n, func):
        super().__init__(func, 2)
        self._field = field
        self._n = n

    @property
    def field(self):
        return self._field

    @property
    def n(self):
        return self._n
    
    def __eq__(self, add2):
        if add2 is self:
            return True
        # Initialize two arbitrary vectors (xs and ys)
        xs, ys = symbols((f'x:{self.n}', f'y:{self.n}'), field=self.field)
        xs, ys = list(xs), list(ys)
        try:
            for lhs, rhs in zip(self.func(xs, ys), add2.func(xs, ys)):
                if not sp.sympify(lhs).equals(sp.sympify(rhs)):
                    return False
            return True
        except Exception:
            return None


class ScalarMul(Operation):
    def __init__(self, field, n, func):
        super().__init__(func, 2)
        self._field = field
        self._n = n

    @property
    def field(self):
        return self._field

    @property
    def n(self):
        return self._n
    
    def __eq__(self, mul2):
        if mul2 is self:
            return True
        # Initialize an arbitrary vector (xs) and scalar (c)
        xs, c = symbols((f'x:{self.n}', 'c'), field=self.field)
        xs = list(xs)
        try:
            for lhs, rhs in zip(self.func(c, xs), mul2.func(c, xs)):
                if not sp.sympify(lhs).equals(sp.sympify(rhs)):
                    return False
            return True
        except Exception:
            return None


class InnerProduct(Operation):
    def __init__(self, func):
        super().__init__(func, 2)


from numbers import Complex, Real
from random import gauss

import sympy as sp

from .mathset import Set
from .parser import split_constraint, sympify
from . import utils as u
from . import vs_utils as vsu


class VectorSpaceError(Exception):
    def __init__(self, msg=''):
        super().__init__(msg)


class NotAVectorSpaceError(Exception):
    def __init__(self, msg=''):
        super().__init__(msg)


class _StandardFn:
    def __init__(self, field, n, constraints=None, *, ns_matrix=None, rs_matrix=None):
        if field not in (Real, Complex):
            raise TypeError('Field must be either Real or Complex.')

        # Verify whether constraints satisfy vector space properties
        if constraints is None:
            constraints = []
        if ns_matrix is None and rs_matrix is None:
            if not is_vectorspace(n, constraints):
                raise NotAVectorSpaceError(
                    'Constraints do not satisfy vector space axioms.'
                    )
        ns, rs = _StandardFn._init_matrices(n, constraints, ns_matrix, rs_matrix)

        self._field = field
        self._n = n
        self._constraints = constraints
        self._ns_matrix = ns
        self._rs_matrix = rs

    @staticmethod
    def _init_matrices(n, constraints, ns_mat, rs_mat):
        if ns_mat is not None:
            ns_mat = sp.zeros(0, n) if u.is_empty(ns_mat) else sp.Matrix(ns_mat)
        if rs_mat is not None:
            rs_mat = sp.zeros(0, n) if u.is_empty(rs_mat) else sp.Matrix(rs_mat)
        
        # Initialize ns_matrix
        if ns_mat is None:
            if rs_mat is None:
                ns_mat = vsu.to_ns_matrix(n, constraints)
            else:
                ns_mat = vsu.to_complement(rs_mat)
        
        # Initialize rs_matrix
        if rs_mat is None:
            rs_mat = vsu.to_complement(ns_mat)
        return ns_mat, rs_mat

    @property
    def field(self):
        return self._field
    
    @property
    def n(self):
        return self._n
    
    @property
    def constraints(self):
        return self._constraints
    
    @property
    def basis(self):
        return self._rs_matrix.tolist()
    
    @property
    def dim(self):
        return len(self.basis)
    
    def __contains__(self, vec):
        if not u.in_field(self.field, *vec):
            return False
        try:
            # Check if vec satisfies vector space constraints
            vec = sp.Matrix(vec)
            return bool((self._ns_matrix @ vec).is_zero_matrix)
        except Exception:
            return False
    
    def __eq__(self, vs2):
        if self is vs2:
            return True
        return self.is_subspace(vs2) and vs2.is_subspace(self)

    # Methods relating to vectors

    def vector(self, std=1, arbitrary=False):
        size = self._rs_matrix.rows
        if arbitrary:
            weights = list(u.symbols(f'c:{size}', field=self.field))
        else:
            weights = [round(gauss(0, std)) for _ in range(size)]
        vec = sp.Matrix([weights]) @ self._rs_matrix
        return vec.flat()  # Return list

    def to_coordinate(self, vector, basis=None):
        if basis is None:
            basis = self._rs_matrix.tolist()
        elif not self._is_basis(*basis):
            raise VectorSpaceError('The provided vectors do not form a basis.')
        if not basis:
            return []
        
        matrix, vec = sp.Matrix(basis).T, sp.Matrix(vector)
        coord_vec = matrix.solve_least_squares(vec)
        return coord_vec.flat()

    def from_coordinate(self, vector, basis=None):  # Check field
        if basis is None:
            basis = self._rs_matrix.tolist()
        elif not self._is_basis(*basis):
            raise VectorSpaceError('The provided vectors do not form a basis.')
        try:
            matrix, coord_vec = sp.Matrix(basis).T, sp.Matrix(vector)
            vec = matrix @ coord_vec
        except Exception as e:
            raise VectorSpaceError('Invalid coordinate vector.') from e
        return vec.flat() if vec else [0] * self.n
    
    def are_independent(self, *vectors):
        matrix = sp.Matrix(vectors)
        return matrix.rank() == matrix.rows
    
    def _is_basis(self, *vectors):
        matrix = sp.Matrix(vectors)
        return matrix.rank() == matrix.rows and len(vectors) == self.dim

    # Methods relating to vector spaces

    def is_subspace(self, vs2):
        for i in range(self._rs_matrix.rows):
            vec = self._rs_matrix.row(i).T
            if not (vs2._ns_matrix @ vec).is_zero_matrix:
                return False
        return True

    # Methods involving the dot product

    def dot(self, vec1, vec2):
        return sum(i * j for i, j in zip(vec1, vec2))
    
    def norm(self, vector):
        return sp.sqrt(self.dot(vector, vector))
    
    def are_orthogonal(self, vec1, vec2):
        return self.dot(vec1, vec2) == 0
    
    def are_orthonormal(self, *vectors):
        # Improve efficiency
        if not all(self.norm(vec) == 1 for vec in vectors):
            return False
        for vec1 in vectors:
            for vec2 in vectors:
                if not (vec1 is vec2 or self.are_orthogonal(vec1, vec2)):
                    return False
        return True
    
    def gram_schmidt(self, *vectors):
        orthonormal_vecs = []
        for v in vectors:
            for q in orthonormal_vecs:
                factor = self.dot(v, q)
                proj = [factor * i for i in q]
                v = [i - j for i, j in zip(v, proj)]
            norm = self.norm(v)
            orthonormal_vecs.append([i / norm for i in v])
        return orthonormal_vecs


class Fn(_StandardFn):
    """
    pass
    """

    def __init__(
            self, field, n, constraints=None, add=None, mul=None, 
            *, isomorphism=None, ns_matrix=None, rs_matrix=None
            ):
        """
        pass
        """
        if constraints is None:
            constraints = []
        if isomorphism is not None:
            if not (isinstance(isomorphism, tuple) and len(isomorphism) == 2):
                raise TypeError('Isomorphism must be a 2-tuple of callables.')
        
        add, mul, iso = Fn._init_operations(field, n, add, mul, isomorphism)

        self._to_standard, self._from_standard = iso
        mapped_constraints = vsu.map_constraints(self._to_standard, constraints)
        super().__init__(
            field, n, mapped_constraints, ns_matrix=ns_matrix, rs_matrix=rs_matrix
            )
        
        self._add = add  # VectorAdd(field, n, add)
        self._mul = mul  # ScalarMul(field, n, mul)
        # Reassign constraints
        self._constraints = constraints

    @staticmethod
    def _init_operations(field, n, add, mul, iso):
        # For efficiency
        if add is None and mul is None:
            iso = (lambda vec: vec, lambda vec: vec)

        if add is None:
            def add(vec1, vec2): return [i + j for i, j in zip(vec1, vec2)]
        if mul is None:
            def mul(scalar, vec): return [scalar * i for i in vec]
        if iso is None:
            iso = vsu.standard_isomorphism(field, n, add, mul)

        return add, mul, iso

    @property
    def add(self):
        return self._add
    
    @property
    def mul(self):
        return self._mul
    
    @property
    def additive_id(self):
        return [0] * self.n
    
    @property
    def additive_inv(self):
        def additive_inv(vec):
            return [-i for i in vec]
        return additive_inv
    
    @property
    def basis(self):
        return [self._from_standard(vec) for vec in super().basis]

    def __contains__(self, vec):
        try:
            standard_vec = self._to_standard(vec)
        except Exception:
            return False
        return super().__contains__(standard_vec)
    
    def __add__(self, vs2):
        return self.sum(vs2)
    
    def __and__(self, vs2):
        return self.intersection(vs2)

    # Methods relating to vectors

    def vector(self, std=1, arbitrary=False):
        standard_vec = super().vector(std, arbitrary)
        return self._from_standard(standard_vec)
    
    def to_coordinate(self, vector, basis=None):
        if basis is not None:
            basis = [self._to_standard(vec) for vec in basis]
        standard_vec = self._to_standard(vector)
        return super().to_coordinate(standard_vec, basis)
    
    def from_coordinate(self, vector, basis=None):
        if basis is not None:
            basis = [self._to_standard(vec) for vec in basis]
        standard_vec = super().from_coordinate(vector, basis)
        return self._from_standard(standard_vec)
    
    def are_independent(self, *vectors):
        standard_vecs = [self._to_standard(vec) for vec in vectors]
        return super().are_independent(*standard_vecs)

    # Methods relating to vector spaces

    def sum(self, vs2):
        rs_matrix = sp.Matrix.vstack(self._rs_matrix, vs2._rs_matrix)
        rs_matrix = u.rref(rs_matrix, remove=True)
        constraints = self.constraints  # Rework
        return Fn(
            self.field, self.n, constraints, self.add, self.mul, 
            isomorphism=(self._to_standard, self._from_standard), 
            rs_matrix=rs_matrix
            )
    
    def intersection(self, vs2):
        ns_matrix = sp.Matrix.vstack(self._ns_matrix, vs2._ns_matrix)
        ns_matrix = u.rref(ns_matrix, remove=True)
        constraints = self.constraints + vs2.constraints
        return Fn(
            self.field, self.n, constraints, self.add, self.mul, 
            isomorphism=(self._to_standard, self._from_standard), 
            ns_matrix=ns_matrix
            )
    
    def span(self, *vectors, basis=None):
        if basis is not None:
            vectors = basis
        standard_vecs = [self._to_standard(vec) for vec in vectors]
        if basis is None:
            standard_vecs = u.rref(standard_vecs, remove=True)
        constraints = [f'span({', '.join(map(str, vectors))})']
        return Fn(
            self.field, self.n, constraints, self.add, self.mul, 
            isomorphism=(self._to_standard, self._from_standard), 
            rs_matrix=standard_vecs
            )

    # Methods involving the dot product
    
    def ortho_complement(self):
        constraints = [f'ortho_complement({', '.join(self.constraints)})']
        return Fn(
            self.field, self.n, constraints, self.add, self.mul, 
            isomorphism=(self._to_standard, self._from_standard), 
            rs_matrix=self._ns_matrix
            )
    
    def ortho_projection(self, vs2):
        raise NotImplementedError()


class VectorSpace:
    """
    pass
    """

    def __init_subclass__(cls, name=None, **kwargs):
        super().__init_subclass__(**kwargs)
        attributes = ['set', 'fn']
        methods = ['__to_fn__', '__from_fn__']

        for attr in attributes:
            if not hasattr(cls, attr):
                raise TypeError(f'{cls.__name__} must define "{attr}".')
        for method in methods:
            if not callable(getattr(cls, method, None)):
                raise TypeError(f'{cls.__name__} must define the method "{method}".')
        
        if not isinstance(cls.set, Set):
            raise TypeError(f'{cls.__name__}.set must be a MathematicalSet.')
        if not isinstance(cls.fn, Fn):
            raise TypeError(f'{cls.__name__}.fn must be of type Fn.')
        if name is not None:
            cls.__name__ = name

    def __init__(self, constraints=None, basis=None, *, fn=None):
        """
        pass
        """
        self.set = Set(self.set.cls, lambda vec: vec in self)
        if fn is not None:
            self.fn = fn
            return
        self.fn = Fn(
            self.fn.field, self.fn.n, constraints, self.fn.add, self.fn.mul, 
            isomorphism=(self.fn._to_standard, self.fn._from_standard)
            )
        if basis is not None:
            if not self.are_independent(*basis):
                raise VectorSpaceError('Basis vectors must be linearly independent.')
            self.fn = self.fn.span(basis=[self.__to_fn__(vec) for vec in basis])
    
    @property
    def field(self):
        """
        {Real, Complex}: The field of scalars.
        """
        return self.fn.field
    
    @property
    def add(self):
        """
        callable: The addition operator on the vector space.
        """
        def add(vec1, vec2):
            fn_vec1, fn_vec2 = self.__to_fn__(vec1), self.__to_fn__(vec2)
            sum = self.fn.add(fn_vec1, fn_vec2)
            return self.__from_fn__(sum)
        return add
    
    @property
    def mul(self):
        """
        callable: The multiplication operator on the vector space.
        """
        def mul(scalar, vec):
            fn_vec = self.__to_fn__(vec)
            prod = self.fn.mul(scalar, fn_vec)
            return self.__from_fn__(prod)
        return mul
    
    @property
    def additive_id(self):
        """
        object: The additive identity of the vector space.
        """
        return self.__from_fn__(self.fn.additive_id)
    
    @property
    def additive_inv(self):
        """
        callable: A function that returns the additive inverse of a given vector.
        """
        def additive_inv(vec):
            fn_vec = self.__to_fn__(vec)
            inv = self.fn.additive_inv(fn_vec)
            return self.__from_fn__(inv)
        return additive_inv
    
    @property
    def basis(self):
        """
        list: The basis of the vector space.
        """
        return [self.__from_fn__(vec) for vec in self.fn.basis]
    
    @property
    def dim(self):
        """
        int: The dimension of the vector space.
        """
        return self.fn.dim

    def __contains__(self, vec):
        """
        Check whether a vector is an element of the vector space.

        Parameters
        ----------
        vec : object
            The vector to check.

        Returns
        -------
        bool
            True if `vec` is an element of `self`, otherwise False.
        """
        if vec not in type(self).set:
            return False
        return self.__to_fn__(vec) in self.fn
    
    def __eq__(self, vs2):
        if self is vs2:
            return True
        return self.is_subspace(vs2) and vs2.is_subspace(self)
    
    def __add__(self, other):
        """
        pass
        """
        if isinstance(other, VectorSpace):
            return self.sum(other)
        return self.coset(other)
    
    def __radd__(self, other):
        return self.coset(other)
    
    def __truediv__(self, vs2):
        """
        Same as ``VectorSpace.quotient``.
        """
        return self.quotient(vs2)
    
    def __and__(self, vs2):
        """
        Same as ``VectorSpace.intersection``.
        """
        return self.intersection(vs2)

    # Methods relating to vectors

    def vector(self, std=1, arbitrary=False):
        """
        Return a vector from the vector space.

        If `arbitrary` is False, then the vector is randomly generated by 
        taking a linear combination of the basis vectors. The weights are 
        sampled from a normal distribution with standard deviation `std`. 
        If `arbitrary` is True, then the general form of the vectors in 
        the vector space is returned.

        Parameters
        ----------
        std : float
            The standard deviation used to generate weights.
        arbitrary : bool, default=False
            Determines whether a random vector or arbitrary vector is returned.
        
        Returns
        -------
        object
            A vector in the vector space.

        Examples
        --------

        >>> vs = fn(Real, 3, constraints=['2*v0 == v1'])
        >>> vs.vector()
        [1, 2, 0]
        >>> vs.vector()
        [-1, -2, 1]
        >>> vs.vector(std=10)
        [11, 22, 13]
        >>> vs.vector(arbitrary=True)
        [c0, 2*c0, c1]
        """
        fn_vec = self.fn.vector(std, arbitrary)
        return self.__from_fn__(fn_vec)
    
    def to_coordinate(self, vector, basis=None):
        """
        Convert a vector to its coordinate vector representation.

        Parameters
        ----------
        vector : object
            A vector in the vector space.
        basis : list, optional
            pass

        Returns
        -------
        list
            The coordinate vector representation of `vector`.

        Raises
        ------
        VectorSpaceError
            If the provided basis vectors do not form a basis for the 
            vector space.

        See Also
        --------
        VectorSpace.from_coordinate

        Examples
        --------

        >>> vs = fn(Real, 3, constraints=['v0 == 2*v1'])
        >>> vs.basis
        [[1, 1/2, 0], [0, 0, 1]]
        >>> vs.to_coordinate([2, 1, 2])
        [2, 0]
        """
        if vector not in self:
            raise TypeError('Vector must be an element of the vector space.')
        if basis is not None:
            if not all(vec in self for vec in basis):
                raise TypeError('Basis vectors must be elements of the vector space.')
            basis = [self.__to_fn__(vec) for vec in basis]

        fn_vec = self.__to_fn__(vector)
        return self.fn.to_coordinate(fn_vec, basis)
    
    def from_coordinate(self, vector, basis=None):
        """
        Convert a coordinate vector to the vector it represents.

        Returns a linear combination of the basis vectors whose weights 
        are given by the coordinates of `vector`. If `basis` is None, then 
        ``self.basis`` is used. The length of `vector` must be equal to 
        the number of vectors in the basis, or equivalently the dimension 
        of the vector space.

        Parameters
        ----------
        vector : list
            The coordinate vector to convert.
        basis : list, optional
            A basis of the vector space.

        Returns
        -------
        object
            The vector represented by `vector`.

        Raises
        ------
        VectorSpaceError
            If `vector` is of incorrect length.

        See Also
        --------
        VectorSpace.to_coordinate

        Examples
        --------

        >>> vs = fn(Real, 3, constraints=['v0 == 2*v1'])
        >>> vs.basis
        [[1, 1/2, 0], [0, 0, 1]]
        >>> vs.from_coordinate([1, 1])
        [1, 1/2, 1]
        >>> new_basis = [[2, 1, 1], [0, 0, 1]]
        >>> vs.from_coordinate([1, 1], basis=new_basis)
        [2, 1, 2]
        """
        if basis is not None:
            if not all(vec in self for vec in basis):
                raise TypeError('Basis vectors must be elements of the vector space.')
            basis = [self.__to_fn__(vec) for vec in basis]
        
        fn_vec = self.fn.from_coordinate(vector, basis)
        return self.__from_fn__(fn_vec)
    
    def are_independent(self, *vectors):
        """
        Check whether the given vectors are linearly independent.

        Returns True if no vectors are given since the empty list is 
        linearly independent by definition.

        Parameters
        ----------
        *vectors
            The vectors in the vector space.

        Returns
        -------
        bool
            True if the vectors are linearly independent, otherwise False.

        Examples
        --------

        >>> vs = fn(Real, 3)
        >>> vs.are_independent([1, 0, 0], [0, 1, 0])
        True
        >>> vs.are_independent([1, 2, 3], [2, 4, 6])
        False
        >>> vs.are_independent([0, 0, 0])
        False
        >>> vs.are_independent()
        True
        """
        if not all(vec in self for vec in vectors):
            raise TypeError('Vectors must be elements of the vector space.')
        fn_vecs = [self.__to_fn__(vec) for vec in vectors]
        return self.fn.are_independent(*fn_vecs)

    # Methods relating to vector spaces

    def sum(self, vs2):
        """
        The sum of two vector spaces.

        Parameters
        ----------
        vs2 : VectorSpace
            The vector space being added.

        Returns
        -------
        VectorSpace
            The sum of `self` and `vs2`.

        Raises
        ------
        TypeError
            If `self` and `vs2` do not share the same ambient space.

        See Also
        --------
        VectorSpace.intersection

        Examples
        --------

        >>> vs1 = fn(Real, 3, constraints=['v0 == v1'])
        >>> vs2 = fn(Real, 3, constraints=['v1 == v2'])
        >>> vs = vs1.sum(vs2)
        >>> vs.basis
        [[1, 0, 0], [0, 1, 0], [0, 0, 1]]
        >>> vs1 + vs2 == vs
        True
        """
        self._validate_type(vs2)
        fn = self.fn.sum(vs2.fn)
        return type(self)(fn=fn)
    
    def intersection(self, vs2):
        """
        The intersection of two vector spaces.

        Parameters
        ----------
        vs2 : VectorSpace
            The vector space to take the intersection with.

        Returns
        -------
        VectorSpace
            The intersection of `self` and `vs2`.

        Raises
        ------
        TypeError
            If `self` and `vs2` do not share the same ambient space.

        See Also
        --------
        VectorSpace.sum

        Examples
        --------

        >>> vs1 = fn(Real, 3, constraints=['v0 == v1'])
        >>> vs2 = fn(Real, 3, constraints=['v1 == v2'])
        >>> vs = vs1.intersection(vs2)
        >>> vs.basis
        [[1, 1, 1]]
        >>> vs1 & vs2 == vs
        True
        """
        self._validate_type(vs2)
        fn = self.fn.intersection(vs2.fn)
        return type(self)(fn=fn)
    
    def span(self, *vectors, basis=None):
        """
        The span of the given vectors.

        Returns the smallest subspace of `self` that contains the vectors 
        in `vectors`. In order to manually set the basis of the resulting 
        space, pass the vectors into `basis` instead. Note that the 
        vectors must be linearly independent if passed into `basis`.

        Parameters
        ----------
        *vectors
            The vectors in the vector space.
        basis : list, optional
            A linearly independent list of vectors in the vector space.

        Returns
        -------
        VectorSpace
            The span of the given vectors.

        Raises
        ------
        VectorSpaceError
            If the provided basis vectors are not linearly independent.

        Examples
        --------

        >>> vs = fn(Real, 3)
        >>> vs.span([1, 2, 3], [4, 5, 6]).basis
        [[1, 0, -1], [0, 1, 2]]
        >>> vs.span(basis=[[1, 2, 3], [4, 5, 6]]).basis
        [[1, 2, 3], [4, 5, 6]]
        >>> vs.span().basis
        []
        """
        if basis is not None:
            return type(self)(basis=basis)
        if not all(vec in self for vec in vectors):
            raise TypeError('Vectors must be elements of the vector space.')
        
        fn_vecs = [self.__to_fn__(vec) for vec in vectors]
        fn = self.fn.span(*fn_vecs)
        return type(self)(fn=fn)
    
    def is_subspace(self, vs2):
        """
        Check whether `self` is a linear subspace of `vs2`.

        Parameters
        ----------
        vs2 : VectorSpace
            The vector space to check.

        Returns
        -------
        bool
            True if `self` is a subspace of `vs2`, otherwise False.

        Examples
        --------

        >>> vs = fn(Real, 3)
        >>> vs1 = fn(Real, 3, constraints=['v0 == v1'])
        >>> vs2 = fn(Real, 3, constraints=['v1 == v2'])
        >>> vs1.is_subspace(vs)
        True
        >>> vs2.is_subspace(vs)
        True
        >>> vs1.is_subspace(vs2)
        False
        >>> vs.is_subspace(vs)
        True
        """
        try:
            self._validate_type(vs2)
        except TypeError:
            return False
        return self.fn.is_subspace(vs2.fn)
    
    # Methods relating to affine spaces
    
    def coset(self, vector):
        """
        pass

        Parameters
        ----------
        vector : object
            A vector in the vector space.

        Returns
        -------
        AffineSpace
            pass

        See Also
        --------
        VectorSpace.quotient
        """
        return AffineSpace(self, vector)
    
    def quotient(self, vs2):
        """
        The quotient of two vector spaces.

        Parameters
        ----------
        vs2 : VectorSpace
            The vector space to divide by.

        Returns
        -------
        VectorSpace
            The quotient of `self` by `vs2`.

        Raises
        ------
        TypeError
            If `vs2` is not a subspace of `self`.

        See Also
        --------
        VectorSpace.coset
        """
        raise NotImplementedError()
        # if not isinstance(vs2, VectorSpace):
        #     raise TypeError()
        # if not vs2.is_subspace(self):
        #     raise TypeError()
        
        # name = f''

        # def in_quotient_space(coset):
        #     return

        # class quotient_space(VectorSpace, name=name):
        #     set = Set(AffineSpace, in_quotient_space, name=name)
        #     fn = Fn(self.field, None, add=self.fn.add, mul=self.fn.mul)
        #     def __to_fn__(self, coset): return
        #     def __from_fn__(self, vec): return
        # return quotient_space()

    # Methods involving the dot product

    def dot(self, vec1, vec2):
        """
        The dot product between two vectors.

        Parameters
        ----------
        vec1, vec2
            The vectors in the vector space.

        Returns
        -------
        float
            The dot product between `vec1` and `vec2`.

        See Also
        --------
        VectorSpace.norm, VectorSpace.are_orthogonal

        Examples
        --------

        >>> vs = fn(Real, 3)
        >>> vs.dot([1, 2, 3], [4, 5, 6])
        32
        >>> vs.dot([1, 0, 1], [0, 1, 0])
        0
        """
        if not (vec1 in self and vec2 in self):
            raise TypeError('Vectors must be elements of the vector space.')
        fn_vec1, fn_vec2 = self.__to_fn__(vec1), self.__to_fn__(vec2)
        return self.fn.dot(fn_vec1, fn_vec2)
    
    def norm(self, vector):
        """
        The norm, or magnitude, of a vector.

        Parameters
        ----------
        vector
            A vector in the vector space.

        Returns
        -------
        float
            The norm of `vector`.

        See Also
        --------
        VectorSpace.dot, VectorSpace.are_orthonormal

        Examples
        --------

        >>> vs = fn(Real, 3)
        >>> vs.norm([1, 2, 3])
        sqrt(14)
        >>> vs.norm([0, 0, 0])
        0
        """
        return sp.sqrt(self.dot(vector, vector))
    
    def are_orthogonal(self, vec1, vec2):
        """
        Check whether two vectors are orthogonal.

        Parameters
        ----------
        vec1, vec2
            The vectors in the vector space.

        Returns
        -------
        bool
            True if the vectors are orthogonal, otherwise False.

        See Also
        --------
        VectorSpace.dot

        Examples
        --------

        >>> vs = fn(Real, 3)
        >>> vs.are_orthogonal([1, 2, 3], [4, 5, 6])
        False
        >>> vs.are_orthogonal([1, 0, 1], [0, 1, 0])
        True
        """
        return self.dot(vec1, vec2) == 0
    
    def are_orthonormal(self, *vectors):
        """
        Check whether the vectors are orthonormal.

        Parameters
        ----------
        *vectors
            The vectors in the vector space.

        Returns
        -------
        bool
            True if the vectors are orthonormal, otherwise False.

        See Also
        --------
        VectorSpace.dot, VectorSpace.norm

        Examples
        --------

        >>> vs = fn(Real, 3)
        >>> vs.are_orthonormal([1, 2, 3], [4, 5, 6])
        False
        >>> vs.are_orthonormal([1, 0, 0], [0, 1, 0])
        True
        >>> vs.are_orthonormal([1, 0, 0])
        True
        >>> vs.are_orthonormal()
        True
        """
        # Improve efficiency
        if not all(self.norm(vec) == 1 for vec in vectors):
            return False
        for vec1 in vectors:
            for vec2 in vectors:
                if not (vec1 is vec2 or self.are_orthogonal(vec1, vec2)):
                    return False
        return True
    
    def gram_schmidt(self, *vectors):
        """
        pass

        Parameters
        ----------
        *vectors
            The vectors in the vector space.

        Returns
        -------
        list
            An orthonormal list of vectors.
        
        Raises
        ------
        VectorSpaceError
            If the provided vectors are not linearly independent.

        See Also
        --------
        VectorSpace.are_orthonormal
        """
        if not self.are_independent(*vectors):
            raise VectorSpaceError('Vectors must be linearly independent.')
        fn_vecs = [self.__to_fn__(vec) for vec in vectors]
        orthonormal_vecs = self.fn.gram_schmidt(*fn_vecs)
        return [self.__from_fn__(vec) for vec in orthonormal_vecs]
    
    def ortho_complement(self):
        """
        The orthogonal complement of a vector space.

        Returns
        -------
        VectorSpace
            The orthogonal complement of `self`.

        See Also
        --------
        VectorSpace.ortho_projection, VectorSpace.dot

        Examples
        --------

        >>> vs = fn(Real, 3, constraints=['v0 == v1'])
        >>> vs.ortho_complement().basis
        [[1, -1, 0]]
        >>> vs.ortho_complement().ortho_complement() == vs
        True
        """
        fn = self.fn.ortho_complement()
        return type(self)(fn=fn)
    
    def ortho_projection(self, vs2):
        """
        The orthogonal projection of `self` onto `vs2`.

        Parameters
        ----------
        vs2 : VectorSpace
            pass

        Returns
        -------
        VectorSpace
            pass

        Raises
        ------
        TypeError
            If `self` and `vs2` do not share the same ambient space.

        See Also
        --------
        VectorSpace.ortho_complement, VectorSpace.dot
        """
        self._validate_type(vs2)
        fn = self.fn.ortho_projection(vs2.fn)
        return type(self)(fn=fn)

    def _validate_type(self, vs2):
        if not isinstance(vs2, VectorSpace):
            raise TypeError()
        if type(self).__name__ != type(vs2).__name__:
            raise TypeError()


class AffineSpace:
    """
    pass
    """

    def __init__(self, vectorspace, vector):
        """
        pass
        """
        if not isinstance(vectorspace, VectorSpace):
            raise TypeError()
        if vector not in type(vectorspace)():
            raise TypeError()
        
        self._vectorspace = vectorspace
        self._representative = vector

    @property
    def vectorspace(self):
        """
        pass
        """
        return self._vectorspace
    
    @property
    def representative(self):
        """
        pass
        """
        return self._representative
    
    @property
    def set(self):
        """
        MathematicalSet: The set containing the points in the affine space.
        """
        return Set(self.vectorspace.set.cls, lambda point: point in self)
    
    @property
    def dim(self):
        """
        int: The dimension of the affine space.
        """
        return self.vectorspace.dim

    def __contains__(self, point):
        """
        Check whether a point is an element of the affine space.

        Parameters
        ----------
        point : object
            The point to check.

        Returns
        -------
        bool
            True if `point` is an element of `self`, otherwise False.
        """
        if point not in type(self.vectorspace)():
            return False
        vec1 = self.representative
        vec2 = self.vectorspace.additive_inv(point)
        return self.vectorspace.add(vec1, vec2) in self.vectorspace

    def __eq__(self, as2):
        if not isinstance(as2, AffineSpace):
            return False
        return self.representative in as2

    def __add__(self, as2):
        if not isinstance(as2, AffineSpace):
            raise TypeError()
        if self.vectorspace != as2.vectorspace:
            raise TypeError('Affine spaces must be cosets of the same vector space.')
        repr = self.vectorspace.add(self.representative, as2.representative)
        return AffineSpace(self.vectorspace, repr)

    def __mul__(self, scalar):
        if not isinstance(scalar, self.vectorspace.field):
            raise TypeError('Scalar must be an element of the field.')
        repr = self.vectorspace.mul(scalar, self.representative)
        return AffineSpace(self.vectorspace, repr)

    def __rmul__(self, scalar):
        return self.__mul__(scalar)


def fn(field, n, constraints=None, basis=None, 
       add=None, mul=None, *, ns_matrix=None, rs_matrix=None):
    """
    pass
    """
    name = f'{field.__name__}^{n}'

    def in_fn(vec):
        try: return sp.Matrix(vec).shape == (n, 1)
        except Exception: return False

    class fn(VectorSpace, name=name):
        set = Set(object, in_fn, name=name)
        fn = Fn(field, n, add=add, mul=mul)
        def __to_fn__(self, vec): return vec
        def __from_fn__(self, vec): return vec

    if not (ns_matrix is None and rs_matrix is None):
        vectorspace = Fn(
            field, n, constraints, add, mul, 
            ns_matrix=ns_matrix, rs_matrix=rs_matrix
            )
        return fn(fn=vectorspace)
    return fn(constraints, basis)


def matrix_space(field, shape, constraints=None, basis=None, 
                 add=None, mul=None):
    """
    pass
    """
    name = f'M({field.__name__}, {shape})'

    def in_matrix_space(mat):
        return mat.shape == shape

    class matrix_space(VectorSpace, name=name):
        set = Set(sp.Matrix, in_matrix_space, name=name)
        fn = Fn(field, sp.prod(shape), add=add, mul=mul)
        def __to_fn__(self, mat): return mat.flat()
        def __from_fn__(self, vec): return sp.Matrix(*shape, vec)
    return matrix_space(constraints, basis)


def poly_space(field, max_degree, constraints=None, basis=None, 
               add=None, mul=None):
    """
    pass
    """
    name = f'P({field.__name__}, {max_degree})'

    def in_poly_space(poly):
        return sp.degree(poly) <= max_degree

    class poly_space(VectorSpace, name=name):
        set = Set(sp.Poly, in_poly_space, name=name)
        fn = Fn(field, max_degree + 1, add=add, mul=mul)
        def __to_fn__(self, poly):
            coeffs = poly.all_coeffs()[::-1]  # Ascending order
            degree_diff = max_degree - len(coeffs) + 1
            return coeffs + ([0] * degree_diff)
        def __from_fn__(self, vec):
            x = sp.symbols('x')
            return sp.Poly.from_list(vec[::-1], x)
    return poly_space(constraints, basis)


def hom(vs1, vs2):
    """
    pass
    """
    if not (isinstance(vs1, VectorSpace) and isinstance(vs2, VectorSpace)):
        raise TypeError()
    if vs1.field is not vs2.field:
        raise VectorSpaceError()
    return matrix_space(vs1.field, (vs2.dim, vs1.dim))


def is_vectorspace(n, constraints):
    """
    Check whether F^n forms a vector space under the given constraints.

    Parameters
    ----------
    n : int
        The length of the vectors in the vector space.
    constraints : list of str
        The constraints ..

    Returns
    -------
    bool
        True if the constraints permit a vector space under standard 
        operations, otherwise False.
    """
    exprs = set()
    for constraint in constraints:
        exprs.update(split_constraint(constraint))
    
    allowed_vars = sp.symbols(f'v:{n}')
    for expr in exprs:
        expr = sympify(expr, allowed_vars)
        if not u.is_linear(expr):
            return False
        
        # Check for nonzero constant terms
        const, _ = expr.as_coeff_add()
        if const != 0:
            return False
    return True


def columnspace(matrix, field=Real):
    """
    Return the column space, or image, of a matrix.

    Parameters
    ----------
    matrix : list of list or sympy.Matrix
        The matrix to take the column space of.
    field : {Real, Complex}
        The field of scalars.

    Returns
    -------
    VectorSpace
        The column space of `matrix`.

    See Also
    --------
    image, rowspace

    Examples
    --------

    >>> matrix = [[1, 2], [3, 4]]
    >>> vs = columnspace(matrix)
    >>> print(vs.basis)
    [[1, 0], [0, 1]]
    >>> vs = image(matrix)
    >>> print(vs.basis)
    [[1, 0], [0, 1]]
    """
    constraints = [f'col({matrix})']
    matrix = sp.Matrix(matrix).T
    matrix = u.rref(matrix, remove=True)
    n = matrix.rows
    return fn(field, n, constraints, rs_matrix=matrix)


def rowspace(matrix, field=Real):
    """
    Return the row space of a matrix.

    Parameters
    ----------
    matrix : list of list or sympy.Matrix
        The matrix to take the row space of.
    field : {Real, Complex}
        The field of scalars.

    Returns
    -------
    VectorSpace
        The row space of `matrix`.

    See Also
    --------
    columnspace

    Examples
    --------

    >>> matrix = [[1, 2], [3, 4]]
    >>> vs = rowspace(matrix)
    >>> print(vs.basis)
    [[1, 0], [0, 1]]
    """
    constraints = [f'row({matrix})']
    matrix = u.rref(matrix, remove=True)
    n = matrix.cols
    return fn(field, n, constraints, rs_matrix=matrix)


def nullspace(matrix, field=Real):
    """
    Return the null space, or kernel, of a matrix.

    Parameters
    ----------
    matrix : list of list or sympy.Matrix
        The matrix to take the null space of.
    field : {Real, Complex}
        The field of scalars.

    Returns
    -------
    VectorSpace
        The null space of `matrix`.

    See Also
    --------
    kernel, left_nullspace

    Examples
    --------

    >>> matrix = [[1, 2], [3, 4]]
    >>> vs = nullspace(matrix)
    >>> print(vs.basis)
    []
    >>> vs = kernel(matrix)
    >>> print(vs.basis)
    []
    """
    constraints = [f'null({matrix})']
    matrix = u.rref(matrix, remove=True)
    n = matrix.cols
    return fn(field, n, constraints, ns_matrix=matrix)


def left_nullspace(matrix, field=Real):
    """
    Return the left null space of a matrix.

    Parameters
    ----------
    matrix : list of list or sympy.Matrix
        The matrix to take the left null space of.
    field : {Real, Complex}
        The field of scalars.

    Returns
    -------
    VectorSpace
        The left null space of `matrix`.

    See Also
    --------
    nullspace

    Examples
    --------

    >>> matrix = [[1, 2], [3, 4]]
    >>> vs = left_nullspace(matrix)
    >>> print(vs.basis)
    []
    >>> matrix = sympy.Matrix([[1, 2], [3, 4]])
    >>> vs1 = left_nullspace(matrix)
    >>> vs2 = nullspace(matrix.T)
    >>> print(vs1 == vs2)
    True
    """
    matrix = sp.Matrix(matrix).T
    return nullspace(matrix, field)


# Aliases
image = columnspace
kernel = nullspace


__version__ = '0.3.0'
__author__ = 'Daniyal Akif'
__email__ = 'daniyalakif@gmail.com'
__license__ = 'MIT'
__description__ = 'A Python package for abstract linear algebra'
__url__ = 'https://github.com/daniyal1249/ablina'


from numbers import Complex, Real

from .innerproduct import InnerProductSpace
from .linearmap import LinearMap, Isomorphism, IdentityMap
from .mathset import MathematicalSet, Set
from .vectorspace import (
    VectorSpace, Fn, AffineSpace, fn, matrix_space, poly_space, hom, 
    is_vectorspace, columnspace, rowspace, nullspace, left_nullspace, 
    image, kernel
    )

__all__ = [
    'Complex', 'Real',
    'InnerProductSpace',
    'LinearMap', 'Isomorphism', 'IdentityMap',
    'MathematicalSet', 'Set',
    'VectorSpace', 'Fn', 'AffineSpace', 'fn', 'matrix_space', 'poly_space',
    'hom', 'is_vectorspace', 'columnspace', 'rowspace', 'nullspace',
    'left_nullspace', 'image', 'kernel'
    ]


import sympy as sp
from sympy.solvers.solveset import NonlinearError

from .parser import ConstraintError, split_constraint, sympify
from .utils import rref, symbols


def additive_id(field, n, add):
    """
    The identity element of an addition function on F^n.

    Parameters
    ----------
    field : {Real, Complex}
        The field of scalars.
    n : int
        The length of the vectors the addition function takes.
    add : callable
        The addition function on F^n.

    Returns
    -------
    pass
    """
    # Initialize an arbitrary vector (xs) and the identity (ys)
    xs, ys = symbols((f'x:{n}', f'y:{n}'), field=field)
    xs, ys = list(xs), list(ys)
    
    # Equations that must be satisfied
    exprs = [sp.expand(lhs - rhs) for lhs, rhs in zip(add(xs, ys), xs)]

    try:
        ids = sp.linsolve(exprs, *ys)
    except NonlinearError:
        ids = sp.nonlinsolve(exprs, ys)  # Check output type
    if isinstance(ids, sp.ConditionSet):
        return []

    valid_ids = []
    for id in ids:
        # Ensure the ids dont depend on xs
        if not any(coord.has(*xs) for coord in id):
            valid_ids.append(list(id))
    return valid_ids


def additive_inv(field, n, add, add_id, lambdify=False):
    """
    The additive inverse of an addition function on F^n.

    Parameters
    ----------
    field : {Real, Complex}
        The field of scalars.
    n : int
        The length of the vectors the addition function takes.
    add : callable
        The addition function on F^n.

    Returns
    -------
    pass
    """
    # Initialize an arbitrary vector (xs) and the inverse (ys)
    xs, ys = symbols((f'x:{n}', f'y:{n}'), field=field)
    xs, ys = list(xs), list(ys)

    # Equations that must be satisfied
    exprs = [sp.expand(lhs - rhs) for lhs, rhs in zip(add(xs, ys), add_id)]

    try:
        inverses = sp.linsolve(exprs, *ys)
    except NonlinearError:
        inverses = sp.nonlinsolve(exprs, ys)
    if isinstance(inverses, sp.ConditionSet):
        return []
    
    if not lambdify:
        return [list(inv) for inv in inverses]

    # Substitute zero for all params if a parametric solution is given
    valid_inverses = []
    sub_zero = {y: 0 for y in ys}
    for inv in inverses:
        valid_inv = []
        for coord in inv:
            valid_inv.append(coord.subs(sub_zero))
        valid_inverses.append(sp.lambdify([xs], valid_inv))
    return valid_inverses


def multiplicative_id(field, n, mul):
    """
    The identity element of a multiplication function on F^n.

    Parameters
    ----------
    field : {Real, Complex}
        The field of scalars.
    n : int
        The length of the vectors the multiplication function takes.
    mul : callable
        The multiplication function on F^n.

    Returns
    -------
    pass
    """
    # Initialize an arbitrary vector (xs) and scalar (c)
    xs, c = symbols((f'x:{n}', 'c'), field=field)
    xs = list(xs)

    # Equations that must be satisfied
    exprs = [lhs - rhs for lhs, rhs in zip(mul(c, xs), xs)]
    ids = sp.nonlinsolve(exprs, [c])  # Check output type
    if isinstance(ids, sp.ConditionSet):
        return []
    
    valid_ids = []
    for id in ids:
        # Ensure the ids dont depend on xs
        if not id[0].has(*xs):
            valid_ids.append(id[0])  # Append scalar instead of tuple
    return valid_ids


def is_commutative(field, n, operation):
    """
    Check whether a binary operation on F^n is commutative.

    Parameters
    ----------
    field : {Real, Complex}
        The field of scalars.
    n : int
        The length of the vectors the operation takes.
    operation : callable
        The operation to check.

    Examples
    --------
    pass
    """
    # Initialize two arbitrary vectors (xs and ys)
    xs, ys = symbols((f'x:{n}', f'y:{n}'), field=field)
    xs, ys = list(xs), list(ys)

    for lhs, rhs in zip(operation(xs, ys), operation(ys, xs)):
        if not sp.sympify(lhs).equals(sp.sympify(rhs)):
            return False
    return True


def is_associative(field, n, operation):
    """
    Check whether a binary operation on F^n is associative.

    Parameters
    ----------
    field : {Real, Complex}
        The field of scalars.
    n : int
        The length of the vectors the operation takes.
    operation : callable
        The operation to check.

    Examples
    --------
    pass
    """
    # Initialize three arbitrary vectors (xs, ys, and zs)
    xs, ys, zs = symbols((f'x:{n}', f'y:{n}', f'z:{n}'), field=field)
    xs, ys, zs = list(xs), list(ys), list(zs)

    lhs_vec = operation(xs, operation(ys, zs))
    rhs_vec = operation(operation(xs, ys), zs)
    for lhs, rhs in zip(lhs_vec, rhs_vec):
        if not sp.sympify(lhs).equals(sp.sympify(rhs)):
            return False
    return True

# To test associativity of multiplication (2 scalars one vector), define
# operation to be normal mul if both are scalars, and scalar mul otherwise

def solve_func_eq(equation, func):
    """
    Attempt to solve a univariate functional equation by guessing common 
    forms of solutions.

    Parameters
    ----------
    equation : sympy.Expr or sympy.Eq
        The functional equation to be solved.
    func : sympy.Function
        The unknown function to solve for.

    Returns
    -------
    valid_funcs : set of sympy.Expr
        pass
    """
    _a, _b, x = sp.symbols('_a _b x')
    w = sp.Wild('w')
    
    solution_forms = [
        lambda x: _a * x + _b,          # Linear
        lambda x: _a * sp.log(x) + _b,  # Logarithmic
        lambda x: _a * sp.exp(x),       # Exponential
    ]
    
    valid_funcs = set()
    for form in solution_forms:
        # Substitute the forms into the equation
        subbed_eq = equation.replace(func(w), form(w))
        invalid_vars = subbed_eq.free_symbols - {_a, _b}
        try:
            sols = sp.solve(subbed_eq, [_a, _b], dict=True)
            sols = sols if sols else [dict()]
        except Exception:
            continue

        for sol in sols:
            invalid_expr = False
            for expr in sol.values():
                if expr.free_symbols.intersection(invalid_vars):
                    invalid_expr = True
                    break
            if invalid_expr:
                continue
            
            if sol or is_tautology(subbed_eq):  # Include tautologies
                valid_func = sp.simplify(form(x).subs(sol))
                valid_funcs.add(valid_func)
    return valid_funcs


def is_tautology(equation):
    """
    Check whether an equation is a tautology.

    Parameters
    ----------
    equation : sympy.Eq
        pass

    Returns
    -------
    bool
        True if `equation` always holds, otherwise False.
    """
    eq = sp.simplify(equation)
    if isinstance(eq, sp.Eq):
        return False
    return bool(eq)  # Must be a sympy bool if not Eq


def standard_isomorphism(field, n, add, mul):
    """
    pass

    Parameters
    ----------
    field : {Real, Complex}
        The field of scalars.
    n : int
        The length of the vectors in the vector space.
    add : callable
        pass
    mul : callable
        pass

    Returns
    -------
    pass
    """
    # Need to support custom domains
    # Need to implement an intersection function
    # Return separate functions for each coordinate

    f = sp.Function('f')
    xs, ys = symbols((f'x:{n}', f'y:{n}'), field=field)

    init_set = False
    for i in range(len(add)):
        func_eq = sp.Eq(f(xs[i]) + f(ys[i]), f(add[i]))
        if not init_set:
            valid_funcs = solve_func_eq(func_eq, f)
            init_set = True
        else:
            valid_funcs.intersection_update(solve_func_eq(func_eq, f))
        if not valid_funcs:
            return valid_funcs
    
    for i in range(len(mul)):
        func_eq = sp.Eq(f(xs[i]) * f(ys[i]), f(mul[i]))
        valid_funcs.intersection_update(solve_func_eq(func_eq, f))
        if not valid_funcs:
            return valid_funcs
    return valid_funcs


# For testing
def standard_isomorphism(field, n, add, mul):
    return lambda x: x, lambda x: x


def map_constraints(mapping, constraints):
    """
    pass

    Parameters
    ----------
    mapping : callable
        pass
    constraints : list of str
        pass

    Returns
    -------
    list of str
        pass
    """
    return constraints


def to_ns_matrix(n, lin_constraints):
    """
    Return the matrix representation of the given linear constraints.

    Parameters
    ----------
    n : int
        pass
    lin_constraints : list of str
        The list of constraints.

    Returns
    -------
    ns_matrix : sympy.Matrix
        A sympy matrix with the linear constraints as rows.
    """
    exprs = set()
    for constraint in lin_constraints:
        exprs.update(split_constraint(constraint))

    matrix = []
    allowed_vars = sp.symbols(f'v:{n}')
    for expr in exprs:
        row = [0] * n
        try:
            expr = sympify(expr, allowed_vars)
        except Exception as e:
            raise ConstraintError('Invalid constraint format.') from e

        for var in expr.free_symbols:
            var_idx = int(var.name.lstrip('v'))
            var_coeff = expr.coeff(var, 1)
            row[var_idx] = var_coeff
        matrix.append(row)
    
    ns_matrix = rref(matrix, remove=True) if matrix else sp.zeros(0, n)
    return ns_matrix


def to_complement(matrix):
    """
    pass

    Parameters
    ----------
    matrix : sympy.Matrix
        pass

    Returns
    -------
    sympy.Matrix
        pass
    """
    if matrix.rows == 0:
        return sp.eye(matrix.cols)
    
    ns_basis = matrix.nullspace()
    if not ns_basis:
        return sp.zeros(0, matrix.cols)
    return rref([vec.T for vec in ns_basis], remove=True)


# Need to account for nested functions using while loop

# x, y, a, b, c = sp.symbols('x y a b c', real=True)
# xs, ys = sp.symbols((f'x:3', f'y:3'), real=True)
# f = sp.Function('f')
# g = sp.Function('g')
# eq = sp.Eq(f(x) * f(y), f(x + y))
# # print(solve_func_eq(eq, f))

# add = [i + j for i, j in zip(xs, ys)]
# mul = [i * j for i, j in zip(xs, ys)]

# print(isomorphism(Real, 3, add, mul))



from .utils import of_arity


class MathematicalSet:
    """
    pass
    """

    def __init__(self, cls, *predicates, name=None):
        """
        pass

        Parameters
        ----------
        cls : type
            The class all set elements must be instances of.
        *predicates
            The predicates all set elements must satisfy.
        name : str, optional
            pass

        Returns
        -------
        MathematicalSet
            pass
        """
        if not isinstance(cls, type):
            raise TypeError()
        if len(predicates) == 1 and isinstance(predicates[0], list):
            predicates = predicates[0]
        if not all(of_arity(pred, 1) for pred in predicates):  # Make sure pred type is valid
            raise ValueError()
        
        self._cls = cls
        self._predicates = remove_duplicates(predicates)
        if name is not None:
            self.__name__ = name

    @property
    def cls(self):
        """
        type: The class that all set elements are instances of.
        """
        return self._cls
    
    @property
    def predicates(self):
        """
        list of callable: The list of predicates all set elements must satisfy.
        """
        return self._predicates
    
    def __repr__(self):
        return (
            f'Set({self.cls.__name__}, '
            f'{[pred.__name__ for pred in self.predicates]})'
            )
    
    def __eq__(self, set2):
        if not isinstance(set2, Set):
            return False
        if hasattr(self, '__name__') and hasattr(set2, '__name__'):
            return self.__name__ == set2.__name__
        # Order of the predicates matters
        return self.cls is set2.cls and self.predicates == set2.predicates
    
    def __contains__(self, obj):
        """
        Check whether an object is an element of the set.

        Parameters
        ----------
        obj : object
            The object to check.

        Returns
        -------
        bool
            True if `obj` is an element of `self`, otherwise False.
        """
        if not isinstance(obj, self.cls):
            return False
        return all(pred(obj) for pred in self.predicates)
    
    def __pos__(self):
        """
        Return `self`.
        """
        return self
    
    def __neg__(self):
        """
        Same as ``Set.complement``.
        """
        return self.complement()
    
    def __and__(self, set2):
        """
        Same as ``Set.intersection``.
        """
        return self.intersection(set2)
    
    def __or__(self, set2):
        """
        Same as ``Set.union``.
        """
        return self.union(set2)
    
    def __sub__(self, set2):
        """
        Same as ``Set.difference``.
        """
        return self.difference(set2)

    def complement(self):
        """
        The complement of a set.

        Returns the set of all objects in the universal set that are not 
        in `self`. The universal set is always ``Set(self.cls)``. In other 
        words, the resulting set contains all instances of ``self.cls`` 
        that are not in `self`.

        Returns
        -------
        MathematicalSet
            The complement of `self`.

        Examples
        --------

        >>> set1 = Set(list, lambda x: len(x) == 3)
        >>> set2 = set1.complement()
        >>> [1, 2, 3] in set1
        True
        >>> [1, 2, 3] in set2
        False
        >>> [1, 2] in set2
        True
        >>> (1, 2) in set2
        False
        >>> None in set2
        False
        """
        def complement_pred(obj):
            return not all(pred(obj) for pred in self.predicates)
        return Set(self.cls, complement_pred)
    
    def intersection(self, set2):
        """
        The intersection of two sets.

        Returns the set of all objects contained in both `self` and 
        `set2`. Note that the ``cls`` attribute of both sets must be the 
        same.

        Parameters
        ----------
        set2 : MathematicalSet
            The set to take the intersection with.

        Returns
        -------
        MathematicalSet
            The intersection of `self` and `set2`.

        Raises
        ------
        ValueError
            If ``self.cls`` and ``set2.cls`` are not the same.

        Examples
        --------

        >>> set1 = Set(list, lambda x: len(x) == 3)
        >>> set2 = Set(list, lambda x: 1 in x)
        >>> set3 = set1.intersection(set2)
        >>> [2, 3, 4] in set3
        False
        >>> [1, 2] in set3
        False
        >>> [1, 2, 3] in set3
        True
        """
        self._validate(set2)
        return Set(self.cls, self.predicates + set2.predicates)

    def union(self, set2):
        """
        The union of two sets.

        Returns the set of all objects contained in either `self` or 
        `set2`. Note that the ``cls`` attribute of both sets must be the 
        same.

        Parameters
        ----------
        set2 : MathematicalSet
            The set to take the union with.

        Returns
        -------
        MathematicalSet
            The union of `self` and `set2`.

        Raises
        ------
        ValueError
            If ``self.cls`` and ``set2.cls`` are not the same.

        Examples
        --------

        >>> set1 = Set(list, lambda x: len(x) == 3)
        >>> set2 = Set(list, lambda x: 1 in x)
        >>> set3 = set1.union(set2)
        >>> [2, 3, 4] in set3
        True
        >>> [1, 2] in set3
        True
        >>> [1, 2, 3] in set3
        True
        """
        self._validate(set2)
        def union_pred(obj):
            return (
                all(pred(obj) for pred in self.predicates) 
                or all(pred(obj) for pred in set2.predicates)
                )
        return Set(self.cls, union_pred)

    def difference(self, set2):
        """
        The difference of two sets.

        Returns the set of all objects in `self` that are not in `set2`.

        Parameters
        ----------
        set2 : MathematicalSet
            The set to be subtracted from `self`.

        Returns
        -------
        MathematicalSet
            The set difference `self` - `set2`.

        Raises
        ------
        ValueError
            If ``self.cls`` and ``set2.cls`` are not the same.

        Examples
        --------

        >>> set1 = Set(list, lambda x: len(x) == 3)
        >>> set2 = Set(list, lambda x: 1 in x)
        >>> set3 = set1.difference(set2)
        >>> [2, 3, 4] in set3
        True
        >>> [1, 2] in set3
        False
        >>> [1, 2, 3] in set3
        False
        """
        return self.intersection(set2.complement())
    
    def is_subset(self, set2):
        """
        Check whether `self` is a subset of `set2`.

        Note that this method is NOT equivalent to the mathematical notion 
        of subset. Due to programmatic limitations, this method instead 
        checks whether every predicate object in `set2` is also in `self`. 
        If so, it returns True, and `self` is a subset of `set2` in the 
        mathematical sense. Otherwise, it returns False, and nothing can 
        be said about the relationship between `self` and `set2`. The 
        examples below illustrate this.

        Parameters
        ----------
        set2 : MathematicalSet
            The set to compare `self` with.

        Returns
        -------
        bool
            True if every predicate in `set2` is also in `self`, 
            otherwise False.

        Raises
        ------
        ValueError
            If ``self.cls`` and ``set2.cls`` are not the same.

        Examples
        --------

        >>> def pred1(x): return len(x) == 3
        >>> def pred2(x): return 1 in x
        >>> set1 = Set(list, pred1)
        >>> set2 = Set(list, pred1, pred2)
        >>> set3 = Set(list, pred1, lambda x: 1 in x)
        >>> set2.is_subset(set1)
        True
        >>> set1.is_subset(set2)
        False
        >>> set3.is_subset(set1)
        True
        >>> set1.is_subset(set3)
        False
        >>> set3.is_subset(set2)
        False
        >>> set2.is_subset(set3)
        False
        """
        self._validate(set2)
        return all(pred in self.predicates for pred in set2.predicates)
    
    def add_predicates(self, *predicates):
        """
        Add predicates to a set.

        Parameters
        ----------
        *predicates
            The predicates to be added.

        Returns
        -------
        MathematicalSet
            `self` with the additional predicates.

        Examples
        --------

        >>> set1 = Set(int, lambda x: x > 0)
        >>> set2 = set1.add_predicates(lambda x: x > 2)
        >>> 1 in set1
        True
        >>> 1 in set2
        False
        >>> 3 in set2
        True
        """
        if len(predicates) == 1 and isinstance(predicates[0], list):
            predicates = predicates[0]
        return Set(self.cls, *self.predicates, *predicates)

    def _validate(self, set2):
        if not isinstance(set2, Set):
            raise TypeError(
                f'Expected a MathematicalSet, got {type(set2).__name__} instead.'
                )
        if self.cls is not set2.cls:
            raise ValueError('The cls attribute of both sets must be the same.')


def remove_duplicates(seq):
    """
    Remove duplicate elements in an iterable while preserving the order.

    Parameters
    ----------
    seq : iterable
        The iterable to remove duplicates from.

    Returns
    -------
    list
        `seq` with duplicates removed.

    Examples
    --------

    >>> lst = [1, 2, 2, 3, 4, 3]
    >>> remove_duplicates(lst)
    [1, 2, 3, 4]
    """
    elems = set()
    return [x for x in seq if not (x in elems or elems.add(x))]


def negate(pred):
    """
    The negation of a predicate.

    Parameters
    ----------
    pred : callable
        The predicate to negate.

    Returns
    -------
    callable:
        The negation of `pred`.

    Examples
    --------

    >>> def pred1(x): return len(x) == 3
    >>> pred2 = negate(pred)
    >>> pred1([1, 2, 3])
    True
    >>> pred2([1, 2, 3])
    False
    """
    def negation(obj): return not pred(obj)
    negation.__name__ = f'not_{pred.__name__}'
    return negation


# Set alias
Set = MathematicalSet


import inspect
from numbers import Complex, Real

import sympy as sp


def symbols(names, field=Complex, **kwargs):
    """
    Returns sympy symbols with the specified names and field (Real or Complex).
    Additional constraints can be specified as keyword args.
    """
    if field is Real:
        return sp.symbols(names, real=True, **kwargs)
    else:
        return sp.symbols(names, **kwargs)


def is_real(expr):
    """
    Returns True if the expression is Real, None if its indeterminate, 
    and False otherwise.
    """
    if hasattr(expr, 'is_real'):
        return expr.is_real  # Sympy's is_real attribute
    return isinstance(expr, Real)


def is_complex(expr):
    """
    Returns True if the expression is Complex, None if its indeterminate, 
    and False otherwise.
    """
    if hasattr(expr, 'is_complex'):
        return expr.is_complex  # Sympy's is_complex attribute
    return isinstance(expr, Complex)


def in_field(field, *exprs):
    """
    Returns True if everything in `exprs` is in the field, otherwise False.
    """
    if field is Real:
        if not all(is_real(expr) for expr in exprs):
            return False
    elif not all(is_complex(expr) for expr in exprs):
        return False
    return True


def is_linear(expr, vars=None):
    """
    Determines if an equation or expression is linear with respect to the 
    variables in vars. If vars is None, all variables in expr are checked.
    """
    if vars is None:
        vars = expr.free_symbols
    if not vars:
        return True
    try:
        terms = sp.Poly(expr, *vars).monoms()
    except sp.PolynomialError:
        return False  # Return False if not a polynomial
    
    for exponents in terms:
        if sum(exponents) not in (0, 1):
            return False
        if not all(i in (0, 1) for i in exponents):
            return False
    return True


def is_empty(matrix):
    """
    Returns True if the matrix contains no elements, otherwise False.
    """
    matrix = sp.Matrix(matrix)
    return matrix.cols == 0 or matrix.rows == 0


def is_invertible(matrix):
    """
    Returns True if the matrix is invertible, otherwise False.
    """
    matrix = sp.Matrix(matrix)
    return matrix.is_square and matrix.det() != 0


def rref(matrix, remove=False):
    """
    Returns the rref of the matrix. Removes all zero rows if remove is True.
    """
    matrix = sp.Matrix(matrix)
    rref, _ = matrix.rref()
    
    if not remove:
        return rref
    for i in range(rref.rows - 1, -1, -1):
        if any(rref.row(i)):
            break
        rref.row_del(i)
    return rref


def of_arity(func, arity):
    """
    Returns True if the function can accept arity positional arguments, 
    otherwise False. Raises a TypeError if func is not callable.
    """
    sig = inspect.signature(func)
    if len(sig.parameters) < arity:
        return False
    
    count_req_pos = 0  # Number of required positional args
    for param in sig.parameters.values():
        # Return False if there are required keyword-only args
        if (param.kind == inspect.Parameter.KEYWORD_ONLY 
            and param.default == inspect.Parameter.empty):
            return False
        if (param.kind in (inspect.Parameter.POSITIONAL_ONLY, 
            inspect.Parameter.POSITIONAL_OR_KEYWORD) 
            and param.default == inspect.Parameter.empty):
            count_req_pos += 1

        if count_req_pos > arity:
            return False
    return True


def add_attributes(cls, *attributes):
    attributes = {attr.__name__: attr for attr in attributes}
    return type(f'{cls.__name__}_subclass', (cls,), attributes)